{-|
Module      : Data.Graph.Clustering.ILouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

ILouvain: really inductive Graph clustering with destructives updates

-}

module Data.Graph.Clustering.ILouvain
  where

import Debug.SimpleReflect
import Data.Set (fromList)
import qualified Data.Set as Set
import Data.Maybe (catMaybes, maybe)
import Data.List (zip, cycle, null)
import Protolude hiding (empty, (&))
import Data.Graph.Inductive
import qualified Data.Graph.Clustering.HLouvain as H

data NodePath = AllNodes | DfsNodes

------------------------------------------------------------------------
-- HyperGraph Definition
type HyperGraph   a b = Gr (Gr () a) b
type HyperContext a b = Context (Gr () a) b
-- TODO Later (hypothesis still)
-- type StreamGraph  a b = Gr a (Gr () b)

------------------------------------------------------------------------
toNodes :: HyperGraph a a -> [[Node]]
toNodes g = map (hnodes g) (nodes g)

isFlat :: HyperGraph a b -> Bool
isFlat g = all (isEmpty . snd) (labNodes g)

------------------------------------------------------------------------
iLouvain :: (Eq a, Show a)
         => Int -> NodePath -> HyperGraph a a -> HyperGraph a a
iLouvain 0 p g = g
iLouvain n p g = trace (show $ toNodes g :: Text) $
  iLouvain' p g (iLouvain (n-1) p g)

iLouvain' :: (Eq a, Show a)
          => NodePath
          -> HyperGraph a a
          -> HyperGraph a a
          -> HyperGraph a a
iLouvain' p g g' = trace (show ps :: Text)
  iLouvain'' g g' ps -- $ filter (\n -> elem n (nodes g)) ps
    where
      -- quick trick to filter path but path of HyperGraph can be different
      ps = if isFlat g'
              then case p of
                     AllNodes -> nodes g
                     DfsNodes -> path' g
              else sortOn (length . hnodes g') (nodes g')

iLouvain'' :: Show a
           => HyperGraph a a
           -> HyperGraph a a
           -> [Node]
           -> HyperGraph a a
iLouvain'' g g' [ ] = g'
iLouvain'' g g' ns  = foldl' (\g1 n -> step g g1 n $ ineighbors g1 n ) g' ns

-- | ineihbors Definition
-- TODO optim sort by jackard
ineighbors :: HyperGraph a a -> Node -> [Node]
ineighbors g n = case isEmpty g of
  True  -> neighbors g n
  False -> filter (\m -> Set.null (Set.intersection (fromList $ hnodes g m) n')
                        && m /= n) (nodes g)
    where
      n'         = H.exclusion candidates (fromList ns)
      candidates = fromList $ concat $ map (neighbors g) ns
      ns         = hnodes g n

step :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> [Node]
     -> HyperGraph a a
step g g' n ns = -- trace ("step'" :: Text) $
  foldl' (\g1 n' -> case match n g1 of
           (Nothing, _) -> g1
           (Just _, _ ) -> step' g g1 n n'
          ) g' ns

step' :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> Node
     -> HyperGraph a a
step' g g' n1 n2 = -- trace ("step " <> show n1 <> " " <> show n2 :: Text) $
  if s2 > 0 && s2 >= s1
  -- if s2 >= s1
    then trace ("step:mv " <> show n1 <> " " <> show n2 :: Text) $
      mv g' [n1] [n2]
    else -- trace ("step:else" :: Text) $
      g'
  where
    s1 = -- trace ("step:mod1" :: Text) $
      imodularity g [n1]
    s2 = -- trace ("step:mod2" :: Text) $
      imodularity g [n1,n2]

stepMax :: Show a
     => HyperGraph a b
     -> HyperGraph a a
     -> Node
     -> [Node]
     -> HyperGraph a a
stepMax g g' n ns = snd
  $ maximumBy (\(n1,g1) (n2,g2) -> compare (imodularity g [n1]) (imodularity g [n2])) gs
    where
      gs = (n, g') : map (\m -> (m, mv g' [n] [m])) ns


------------------------------------------------------------------------
hnodes :: HyperGraph a b -> Node -> [Node]
hnodes g n = case match n g of
  (Nothing          , _) -> []
  (Just (p, n, l, s), _) -> n : nodes l
{-
hdeg :: Graph gr => gr a b -> Node -> Maybe Int
hdeg = undefined
-}
------------------------------------------------------------------------
-- TODO go depth in HyperGraph (modularity at level/depth)
imodularity :: HyperGraph a b -> [Node] -> Double
imodularity g ns = -- trace ("imodul" :: Text) $
                   H.hmodularity g
                 $ fromList
                 $ concat
                 $ map (hnodes g) ns

------------------------------------------------------------------------
toHyperGraph :: Gr () Double -> HyperGraph Double Double
toHyperGraph g = nmap (\_ -> empty) g
------------------------------------------------------------------------
-- Spoon Graph
--    1
--   / \
--  2   3
--   \ /
--    4
--    |
--    5
spoon :: HyperGraph Double Double
spoon = mkGraph ns es
  where
    ns :: [LNode (Gr () Double)]
    ns = zip [1..6] (cycle [empty])

    es :: [LEdge Double]
    es = [ (1, 2, 1.0)
         , (1, 3, 1.0)
         , (2, 4, 1.0)
         , (3, 4, 1.0)
         , (4, 5, 1.0)
         ]

-- | Needed function
-- mv: a Node elsewhere in the HyperGraph

-- Move Properties:
-- mv Node into another Node and remove it again should be equal to
-- identity

-- property test:
-- mv (mv g [a] [b]) [b,a] [] = identity
-- mv g [a] [] == g
-- mv g [a,b] [] /= g
-- mv (mv spoon [1] [4]) [4,1] [] = identity

-- Move target type
-- let's start simple: path lenght <= 2 max
mv :: Show a => HyperGraph a a
   -> [Node] -> [Node]
   -> HyperGraph a a
mv g [ ] [ ] = g
mv g [_] [ ] = g
mv g [ ] [_] = g

mv g [a] [b] = case a == b of
  True  -> panic "mv, impossible: moved node is same as destination"
  False -> mv' g a b
mv g [a,b] [ ] = case match a g of
  (Nothing,  _) -> panic $ "mv: fst Node of Path does not exist: " <> show a
  (Just (p, n, l, s), g1) -> case match b l of
    (Just (p',n',l',s'), g2)  -> (p', n', g2, s')
                               & ((p , n , delNode b l , s )
                               & g1)
    (Nothing, _) -> panic "mv: snd Node of Path does not exist"

mv g (x:xs) (y:ys) = panic "mv: path too long"

                  ----------------------------

-- | Start simple (without path)
mv' :: Show a => HyperGraph a a
    -> Node -> Node
    -> HyperGraph a a
mv' g n1 n2 = -- trace (show (c1,c2) :: Text) $ 
  (mvMContext c1 c2) & g2
    where
      (c1, g1) = match n1 g
      (c2, g2) = match n2 g1

mvMContext :: Maybe (HyperContext a a)
           -> Maybe (HyperContext a a)
           -> HyperContext a a
mvMContext (Just (a1 ,n ,l ,a2 ))
           (Just (a1',n',l',a2')) = (a1',n', merge l' (c & l),a2')
            where
              c = (a1, n, (), a2)

mvMContext _ (Just _) = panic "mvmcontext: First Node  does not exist"
mvMContext (Just _) _ = panic "mvmcontext: Snd   Node  does not exist"
mvMContext _ _        = panic "mvmcontext: Both  Nodes do   not exist"

merge :: (Graph gr, DynGraph gr)
      => gr a b -> gr a b -> gr a b
merge = ufold (&)

------------------------------------------------------------------------
test_mv :: (Ord a, Show a) => HyperGraph a a -> Node -> Node -> Bool
test_mv g a b = (mv (mv g [a] [b]) [b,a] []) == g

------------------------------------------------------------------------
-- Paths in the Graph to be tested
-- Directed graph strategy
path_dir :: Graph gr => gr a b -> [[Node]]
path_dir g = map (\xs -> dfs xs g ) (components g)

-- UnDirected graph strategy
path :: (DynGraph gr, Eq b) => gr a b -> [[Node]]
path g' = map sortNodes cs
  where
    sortNodes ns = case head $ sortOn (Down . (deg g)) ns of
      Nothing -> []
      Just  n -> dfs [n] g  -- dfs for glustering, bfs for klustering
    cs = components g
    g  = undir g'

path' :: (DynGraph gr, Eq b) => gr a b -> [Node]
path' = maybe [] identity  . head . path

------------------------------------------------------------------------
------------------------------------------------------------------------
